/*
 *  USBSink - dirtransfer.cc
 *
 *  Copyright (C) 2006-2007 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "dirtransfer.h"
#include "logger.h"
#include "task.h"
#include "utils.h"

#include <list>

#include <glib/gi18n.h>

#include <glibmm.h>

#include <libgnomevfs/gnome-vfs-result.h>
#include <libgnomevfsmm/uri.h>

using std::list;

namespace USBSink
{

namespace Transfer
{

typedef list<Glib::ustring> ustring_list;
typedef list<Glib::ustring>::iterator ustring_list_iterator;

typedef list< Glib::RefPtr<const Gnome::Vfs::Uri> > uri_list;

typedef Glib::ListHandle< Glib::RefPtr<const Gnome::Vfs::Uri> > ListHandleUris;

DirTransfer::DirTransfer(const Glib::ustring& root_src_path,
                         const Glib::ustring& root_dest_path)
{
  root_src_path_ = Util::String::delete_trailing_separator(root_src_path);
  root_dest_path_ = Glib::build_filename(
    Util::String::delete_trailing_separator(root_dest_path),
    Glib::path_get_basename(root_src_path));
}

DirTransfer::~DirTransfer()
{}

void
DirTransfer::add_file(const Glib::ustring& _path)
{
  Glib::ustring path(Util::String::delete_trailing_separator(_path));
  src_files_.push_back(path);
}

Glib::ustring
DirTransfer::build_destination_path(const Glib::ustring& source_file)
{
  Glib::ustring dest_path;

  if (source_file.find(root_src_path_) != Glib::ustring::npos)
    dest_path =
      Glib::build_filename(
	root_dest_path_, source_file.substr(
	  root_src_path_.length(), source_file.length()));

  return dest_path;
}

void
DirTransfer::run(Logger* logger,
                 TaskHandling::Task* task,
                 GnomeVFSAsyncXferProgressCallback callback)
{
  ustring_list source_path_list;
  ustring_list target_path_list;

  ustring_list_iterator it(src_files_.begin());
  ustring_list_iterator end(src_files_.end());

  for ( ; it != end; ++it)
  {
    Glib::ustring target = build_destination_path(*it);
    bool needs_copy = false;
    
    if (Glib::file_test(target, Glib::FILE_TEST_EXISTS))
    {
      const Glib::RefPtr<Gnome::Vfs::Uri> src_uri =
	Gnome::Vfs::Uri::create(*it);
      const Glib::RefPtr<Gnome::Vfs::Uri> target_uri =
	Gnome::Vfs::Uri::create(target);
      
      if (src_uri->get_file_info()->get_modification_time() >
	  target_uri->get_file_info()->get_modification_time())
	needs_copy = true;
    }
    else
      needs_copy = true;

    if (needs_copy && (! Glib::file_test(*it, Glib::FILE_TEST_IS_DIR)))
    {
      source_path_list.push_back(*it);
      target_path_list.push_back(target);

      if (! Glib::file_test(Glib::path_get_dirname(target),
			    Glib::FILE_TEST_EXISTS))
	Util::File::make_missing_dirs_for_file(target);
    }
  }

  if (! source_path_list.empty())
  {
    // TODO: there's too much copying here. Refactor to use DirTransfer
    // with Uris, not path strings.

    uri_list source_uris, target_uris;

    ustring_list_iterator sources_it = source_path_list.begin();
    ustring_list_iterator sources_end = source_path_list.end();

    *logger << Util::i18n_printf(_("%s: Preparing to copy files:"),
                                 Util::Time::get_current_time().c_str())
            << "\n";

    for ( ; sources_it != sources_end; ++sources_it)
    {
      *logger << *sources_it << "\n";
      source_uris.push_back(Gnome::Vfs::Uri::create(*sources_it));
    }

    *logger << Util::i18n_printf(_("Target directory is %s"),
                                 root_dest_path_.c_str())
            << "\n";

    ustring_list_iterator targets_it = target_path_list.begin();
    ustring_list_iterator targets_end = target_path_list.end();

    for ( ; targets_it != targets_end; ++targets_it)
    {
      target_uris.push_back(Gnome::Vfs::Uri::create(*targets_it));
    }

    //
    ListHandleUris source_proxy(source_uris);
    ListHandleUris target_proxy(target_uris);
    GnomeVFSResult xfer_result;
    CallbackData* callback_data = new CallbackData;
    
    callback_data->logger = logger;
    callback_data->task = task;

    task->inc_transfer_count();

    xfer_result = gnome_vfs_async_xfer(&async_handle_,
                                       source_proxy.data(),
                                       target_proxy.data(),
                                       GNOME_VFS_XFER_DEFAULT,
                                       GNOME_VFS_XFER_ERROR_MODE_QUERY,
                                       GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
                                       GNOME_VFS_PRIORITY_DEFAULT,
                                       callback,
                                       callback_data,
                                       NULL,   /* progress_sync_callback */
                                       NULL);  /* sync_callback_data */

    if (xfer_result != GNOME_VFS_OK)
    {
      *logger << Util::i18n_printf(_("Transfer failed: %s."),
                                   gnome_vfs_result_to_string(xfer_result))
              << "\n";

      g_warning("gnome-vfs transfer failed: %s",
                gnome_vfs_result_to_string(xfer_result));

      task->dec_transfer_count();
    }

    // TODO: wait for libgnomevfsmm release with async2 and transfer2.
  }
}

} // namespace Transfer

} // namespace USBSink
