/*
 *  USBSink - filelistbase.cc
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "filelistbase.h"
#include "filelistitem.h"
#include "filelistloader.h"
#include "globalstrings.h"
#include "utils.h"

#include <algorithm>
#include <list>
#include <vector>

#include <glibmm/miscutils.h>

using std::list;
using std::vector;

namespace USBSink
{

namespace TaskHandling
{

void
FileListBase::init(const Glib::ustring& name)
{
  name_ = name;
  path_ = Glib::build_filename(Util::Config::get_file_list_dir(), name);
  format_str_ = Util::Config::get_file_list_format_string(); // default
}

/**
 * @param name A name for the list, used to construct the path where the list
 * will be saved using Util::Config::get_file_list_dir().
 */
FileListBase::FileListBase(const Glib::ustring& name)
{
  init(name);
}

FileListBase::FileListBase(const Glib::ustring& name,
                           const Glib::RefPtr<Gtk::TextBuffer>& buffer,
                           const ClientSlot& slot)
{
  init(name);

  list<SharedItem> items;

  FileListLoader::get_items(buffer, slot, items);

  list<SharedItem>::iterator it(items.begin());
  list<SharedItem>::iterator end(items.end());

  for ( ; it != end; ++it)
    add(*it);
}

FileListBase::~FileListBase()
{
}

/**
 * Adds a new Item into the list.
 *
 * @param new_item An Item to add into the list.
 */
void
FileListBase::add(SharedItem& new_item)
{
  items_.push_back(new_item);
}

/**
 * @return the name of the list formatted using a format string specified via
 * the set_format_string() function. If a custom format has not been set,
 * a default one will be used.
 */
Glib::ustring
FileListBase::get_name_formatted() const
{
  GString* gstr;
  Glib::ustring result;

  gstr = g_string_new("");
  g_string_printf(gstr, format_str_.c_str(), name_.c_str());

  result = gstr->str;

  g_string_free(gstr, TRUE);

  return result;
}

/**
 * @return a list of formatted names of each item in the FileList, as returned
 * by get_name_formatted().
 */
vector<Glib::ustring>
FileListBase::get_text_for_display() const
{
  vector<Glib::ustring> result(items_.size());

  list<SharedItem>::const_iterator it(items_.begin());
  list<SharedItem>::const_iterator end(items_.end());

  for ( ; it != end; ++it)
     result.push_back((*it)->get_name_formatted());

  return result;
}

Glib::ustring
FileListBase::get_text_string_for_display() const
{
  Glib::ustring contents;
  vector<Glib::ustring> file_list_contents(get_text_for_display());

  vector<Glib::ustring>::const_iterator fl_it(file_list_contents.begin());
  vector<Glib::ustring>::const_iterator fl_end(file_list_contents.end());

  for ( ; fl_it != fl_end; ++fl_it)
  {
    if (*fl_it != Glib::ustring(""))
    {
      contents += ((*fl_it) + '\n');
    }
  }

  return contents;
}

/**
 * @param str UTF-8 string to match.
 *
 * @return true if the specified string matches any FileListItem contained.
 */
bool
FileListBase::match(const Glib::ustring& str) const
{
  bool result = false;

  list<SharedItem>::const_iterator it(items_.begin());
  list<SharedItem>::const_iterator end(items_.end());
    
  for ( ; it != end; ++it)
  {
    if ((*it)->match(str))
    {
      result = true;
      break;
    }
  }

  return result;
}

/** Assigns new items to the list. All previous are first cleared.
 *
 * @param new_items New items that should make up the list.
 */
void
FileListBase::set_items(const list<SharedItem>& new_items)
{
  items_.clear();
  items_.assign(new_items.begin(), new_items.end());
}

/**
 * @param name Name of the item to search for.
 *
 * @return Whether the list contains it.
 */
bool
FileListBase::contains_item(const Glib::ustring& name) const
{
  bool result = false;

  list<SharedItem>::const_iterator it = items_.begin();
  list<SharedItem>::const_iterator end = items_.end();

  for ( ; it != end; ++it)
  {
    if ((*it)->get_name() == name)
    {
      result = true;
      break;
    }
  }

  return result;
}

/**
 * @param item Item to which should be replace and old object.
 */
void
FileListBase::reset_item(SharedItem& item)
{
  list<SharedItem>::iterator it = items_.begin();
  list<SharedItem>::iterator end = items_.end();

  for ( ; it != end; ++it)
  {
    if ((*it)->get_name() == item->get_name())
    {
      break;
    }
  }

  if (it != end)
  {
    items_.erase(remove(items_.begin(), items_.end(), *it), items_.end());
    items_.push_back(item);
  }
}

} // namespace TaskHandling

} // namespace USBSink
