// -*- c++ -*-
/*
 *  USBSink - filelist.h
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _TASK_FILE_LIST_BASE_H_
#define _TASK_FILE_LIST_BASE_H_

#include "exception.h"
#include "filelist.h"
#include "clientdispatcher.h"
#include "filelistitem.h"
#include "taskitem.h"
#include "uncopyable.h"

#include <list>
#include <tr1/memory>

#include <gtkmm/textbuffer.h>

namespace USBSink
{

namespace TaskHandling
{

typedef std::tr1::shared_ptr<Item> SharedItem;

/** Thrown at construction of FileList objects.
 * It is meant to indicate major errors as items are being pushed into a file
 * list, or a new list is being created, such as bad regular expressions
 * or non-existing external lists.
 */
class FileListException : public Exception
{
 public:

  FileListException() {}

  std::list<Glib::ustring> get_invalid_regexps() const;
  std::list<Glib::ustring> get_invalid_list_paths() const;

  void add_invalid_item(const Glib::ustring& val);
  void add_invalid_list_paths(const Glib::ustring& val);

 private:

  std::list<Glib::ustring> invalid_items_;
  std::list<Glib::ustring> invalid_list_paths_;
};

/** A list that defines task policy - which files to include or ignore in
 * a synchronization process.
 *
 * ...
 *
 * Client code is not supposed to create and delete any type of FileList
 * directly, since lists are generally shared between tasks. Factory methods
 * of FileListRegistry should be used instead.
 */
class FileListBase : public FileList
{
 public:

  Glib::ustring get_name() const;
  Glib::ustring get_name_formatted() const;
  Glib::ustring get_format_string() const;

  Glib::ustring get_path() const;

  std::vector<Glib::ustring> get_text_for_display() const;
  Glib::ustring              get_text_string_for_display() const;

  void set_format_string(const Glib::ustring& format);

  virtual bool match(const Glib::ustring& str) const;

  virtual void set_items(const std::list<SharedItem>& items);

  virtual bool contains_item(const Glib::ustring& name) const;

  virtual void reset_item(SharedItem& item);

 protected:

  explicit FileListBase(const Glib::ustring& name);
  explicit FileListBase(const Glib::ustring& name,
                        const Glib::RefPtr<Gtk::TextBuffer>& buffer,
                        const ClientSlot& slot);

  virtual ~FileListBase();

  virtual void add(SharedItem& item);

  std::list<SharedItem>::const_iterator begin() const;
  std::list<SharedItem>::const_iterator end() const;

  void set_name(const Glib::ustring& ) {} // no implementation

 private:

  void init(const Glib::ustring& name);

  Glib::ustring name_;
  Glib::ustring path_;
  Glib::ustring format_str_;

  std::list<SharedItem> items_;
};


/* Implementation of FileListException's inline functions. */

/** @return a list of invalid UTF-8 regular expression strings. */
inline std::list<Glib::ustring>
FileListException::get_invalid_regexps() const
{
  return invalid_items_;
}

/** @return a list of invalid paths to file lists. */
inline std::list<Glib::ustring>
FileListException::get_invalid_list_paths() const
{
  return invalid_list_paths_;
}

inline void
FileListException::add_invalid_item(const Glib::ustring& val)
{
  invalid_items_.push_back(val);
}

inline void
FileListException::add_invalid_list_paths(const Glib::ustring& val)
{
  invalid_list_paths_.push_back(val);
}

/* Implementation of FileList's inline functions. */

/** @return path to file list on the filesystem. */
inline Glib::ustring
FileListBase::get_name() const
{
  return name_;
}

/**
 * @return C-style format string used to produce the result of
 * get_name_formatted().
 */
inline Glib::ustring
FileListBase::get_format_string() const
{
  return format_str_;
}

inline Glib::ustring
FileListBase::get_path() const
{
  return path_;
}

/**
 * @param new_format C-style format string used to produce the result of
 * get_name_formatted().
 */
inline void
FileListBase::set_format_string(const Glib::ustring& new_format)
{
  format_str_ = new_format;
}

/**
 * @return const_iterator to iterate over the list's items.
 */
inline std::list<SharedItem>::const_iterator
FileListBase::begin() const
{
  return items_.begin();
}

/**
 * @return const_iterator of the end of the list's items.
 */
inline std::list<SharedItem>::const_iterator
FileListBase::end() const
{
  return items_.end();
}

} // namespace TaskHandling

} // namespace USBSink

#endif // _TASK_FILE_LIST_BASE_H_
