/*
 *  USBSink - filelistloader.cc
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "clientdispatcher.h"
#include "filelistitem.h"
#include "filelistloader.h"
#include "filelistregistry.h"
#include "globalstrings.h"
#include "utils.h"
#include "waitingclient.h"

#include <string>
#include <tr1/memory>

#include <pcrecpp.h>

using std::list;

typedef list<Glib::ustring> ustring_list;

namespace USBSink
{

namespace TaskHandling
{

void
FileListLoader::get_items(const Glib::RefPtr<Gtk::TextBuffer>& buffer,
                          list<SharedItem>& items)
{
  ClientSlot* slot = 0;
  get_items_private(buffer, slot, items);
}

void
FileListLoader::get_items(const Glib::RefPtr<Gtk::TextBuffer>& buffer,
                          const ClientSlot& slot,
                          list<SharedItem>& items)
{
  get_items_private(buffer, const_cast<ClientSlot*>(&slot), items);
}

void
FileListLoader::get_items_private(const Glib::RefPtr<Gtk::TextBuffer>& buffer,
                                  ClientSlot* slot,
                                  list<SharedItem>& items)
{
  ustring_list lines;
  Glib::ustring regexp_string;
  std::string match;

  // Build the regular expression to extract a file list name from
  // formatted version:
  regexp_string = "^";
  regexp_string += file_list_format_left_terminator;
  regexp_string += "(.*)";
  regexp_string += file_list_format_right_terminator;
  regexp_string += "$";

  pcrecpp::RE regexp(regexp_string.c_str(), pcrecpp::UTF8());

  // Get all lines as ustrings, iterate:
  lines = Util::String::get_lines_from_text_buffer(buffer);

  ustring_list::iterator lines_it(lines.begin());
  ustring_list::iterator lines_end(lines.end());

  for ( ; lines_it != lines_end; ++lines_it)
  {
    if (regexp.FullMatch(lines_it->c_str(), &match))
    {
      // match holds the list name,
      // now try to get the list from the registry:
      try
      {
        if (! FileListRegistry::is_initialized())
        {
          if (slot)
          {
            ClientDispatcher* dispatcher = ClientDispatcher::instance();
            SharedWaitingClient client(new WaitingClient(*slot));
            client->add_request(match);
            dispatcher->add_client(client);
          }
        }
        else
        {
          FileListRegistry* registry = FileListRegistry::instance();
          SharedItem lst(registry->get_file_list(match));
          items.push_back(lst);
        }
      }
      catch (FileListNotFoundException& ex)
      {
	//TODO: figure out how to pass this to UI
        g_warning ("File list '%s' not found", match.c_str());
      }
    }
    else
    {
      // It's a regular expression - create an item and go on:
      if (Util::Regexp::validate(*lines_it))
      {
	SharedFileListItem item = FileListItem::create(*lines_it);
	items.push_back(item);
      }
    }
  }
}

} // namespace TaskHandling

} // namespace USBSink
