/*
 *  USBSink - filewatcher.cc
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "filewatcher.h"

#include <fam.h>
#include <unistd.h>
#include <fcntl.h>

#include <algorithm>
#include <memory>
#include <map>
#include <utility>

#include <tr1/memory>

#include <glib/gerror.h>
#include <glib/gmessages.h>

#include <glibmm/fileutils.h>
#include <glibmm/main.h>
#include <glibmm/miscutils.h>

using std::auto_ptr;
using std::map;
using std::pair;

using std::tr1::shared_ptr;

namespace USBSink
{

auto_ptr<FileWatcher> FileWatcher::instance_;

FileWatcher*
FileWatcher::instance()
{
  if (! instance_.get())
    instance_.reset(new FileWatcher);

  return instance_.get();
}

bool
FileWatcher::open_fam()
{
  // Open a connection to FAM, introduce ourselves.
  if (FAMOpen2(&fam_, "usbsink") != 0)
    return false;

  // Create a new FAM file descriptor and set the connection
  // to be non-blocking. This is taken from SGI's online document.
  int fam_fd = FAMCONNECTION_GETFD(&fam_);
  int flags = fcntl(fam_fd, F_GETFL);
    
  if (! (flags & O_NONBLOCK))
    fcntl(fam_fd, F_SETFL, flags | O_NONBLOCK);

  // Set Glib to monitor the FAM file descriptor.
  // The effect is that we have a signal handler for FAM events.
  fam_signal_connection_ = Glib::signal_io().connect(
    sigc::mem_fun(*this, &FileWatcher::on_fam_event),
    fam_fd,
    Glib::IO_IN);

  started_ = true;

  return true;
}

/**
 * Creates a FAM connection so that monitoring starts immediately
 * as a file or directory is added.
 */
FileWatcher::FileWatcher()
  :
  started_(false)
{
  if (! open_fam())
  {
    g_warning("Can't connect to FAM; is gamin installed or famd running?\n");
    throw FamError("Error while trying to connect to FAM.");
  }
}

void
FileWatcher::close_fam()
{
  if (! started_) // extreme case, complain again
  {
    g_warning("Closing FileWatcher: a FAM connection was not opened.");
    return;
  }

  // Disconnect the handler from the main loop, close the connection to FAM.
  fam_signal_connection_.disconnect();
  FAMClose(&fam_);
  started_ = false;
}

/**
 * Stops monitoring and cleans up used resources.
 */
FileWatcher::~FileWatcher()
{
  remove_all_entries();
  close_fam();
}

/**
 * Utility function for getting an Entry for a given path.
 */
shared_ptr<FileWatcher::Entry>
FileWatcher::get_entry(const Glib::ustring& _path)
{
  if (! Glib::path_is_absolute(_path))
  {
    Entry* pe = 0;
    shared_ptr<Entry> e(pe);
    return e;
  }

  Glib::ustring path(_path);

  // We want the string path to be in format "/home/name/some_dir".
  if ((! path.empty()) && (path[path.size() - 1] == '/'))
    path = path.substr(0, path.length() - 1);

  EntryMap::iterator it = files_to_monitor_.find(path);

  if (it == files_to_monitor_.end())
  {
    Entry* pe = 0;
    shared_ptr<Entry> e(pe);
    return e;
  }
  else // all ok
    return it->second;
}

/**
 * The function that processes any pending fam events.
 */
bool
FileWatcher::on_fam_event(Glib::IOCondition /* condition */)
{
  while (started_ && FAMPending(&fam_)) // there is an event waiting
  {
    FAMEvent event; // the struct where changes to files/dirs are encoded

    if (FAMNextEvent(&fam_, &event) != 1)
    {
      // A serious problem.
      close_fam();
      g_error("Error trying to process FAM event, connection closed.");
      return false; // disconnect the handler
    }

    if ((event.code == FAMChanged) ||
	(event.code == FAMDeleted) ||
	(event.code == FAMCreated) ||
	(event.code == FAMMoved))
    {
      // Find the entry that corresponds to event's request number.
      shared_ptr<Entry> e;
      EntryMap::iterator it(files_to_monitor_.begin());
      EntryMap::iterator end(files_to_monitor_.end());

      for ( ; it != end; ++it)
      {
	if (FAMREQUEST_GETREQNUM(&(it->second->get_request())) ==
	    FAMREQUEST_GETREQNUM(&(event.fr)))
	{
	  e = it->second; // found it
	  break;
	}
      }
      
      if (it == end)
	return true; // should never happen :)
      
      e->set_dirty(true);
      /*g_debug("\n\nFW: handling an event for request %i...",
	FAMREQUEST_GETREQNUM(&event.fr));*/

      if (e->is_dir())
      {
	switch (event.code)
	{
	case FAMChanged:
	{
	  /* g_debug("FAM event: file '%s' in '%s' modified, req# %i\n",
		  event.filename,
		  e->get_path().c_str(),
		  FAMREQUEST_GETREQNUM(&event.fr));*/

	  //e->set_change_news(FILE_STATUS_CHANGED);
	  Glib::ustring changed_file(Glib::build_filename(e->get_path(),
							  event.filename));
	  signal_monitor(e->get_path(), changed_file, FILE_STATUS_CHANGED);
	}
	  break;

	case FAMDeleted:
	  /*g_debug("FAM event: '%s' deleted from '%s', req# %i\n",
		  event.filename,
		  e->get_path().c_str(),
		  FAMREQUEST_GETREQNUM(&event.fr));*/

	  // Check once again: all watched dirs should have absolute paths.
	  if (Glib::path_is_absolute(e->get_path()))
	  {
	    if (e->get_path() == event.filename)
	    {
              // The monitored directory has been deleted.
	      e->set_change_news(FILE_STATUS_DELETED);
	      FAMCancelMonitor(&fam_, &event.fr);
	    }
	    else
	    {
	      // A file or directory inside has been deleted.
	      Glib::ustring path(Glib::build_filename(e->get_path(),
						      event.filename));
	      shared_ptr<Entry> deleted_entry(get_entry(path));

	      if (deleted_entry.get()) // it is being monitored
	      {
		deleted_entry->set_change_news(FILE_STATUS_DELETED);
	      }
	    }
	    signal_monitor.emit(e->get_path(),
				event.filename,
				FILE_STATUS_DELETED);
	  }
	  break;
	
	case FAMCreated:
	{ // the braces must be here to prevent compiler error regarding
	  // created_file being without scope

	  // A new file or directory has been created in a monitored directory.
	  /*g_debug("FAM event: '%s' created in '%s', req# %i\n",
		  event.filename,
		  e->get_path().c_str(),
		  FAMREQUEST_GETREQNUM(&event.fr));*/

	  Glib::ustring created_file(Glib::build_filename(e->get_path(),
							  event.filename));
	  // Don't add the created file here - a task which handles the signal
	  // will decide.
	  signal_monitor.emit(e->get_path(),
			      created_file,
			      FILE_STATUS_CREATED);
	}
	  break;

	case FAMStartExecuting:
	case FAMStopExecuting:
	case FAMMoved:
	case FAMAcknowledge:
	case FAMExists:
	case FAMEndExist:
	  break;	
	}
      }
      else // an event regarding a regular file
      {
	switch (event.code)
	{
	case FAMChanged:
	  /*g_debug("FAM event: file '%s' modified, req# %i\n",
		  e->get_path().c_str(),
		  FAMREQUEST_GETREQNUM(&event.fr));*/

	  e->set_change_news(FILE_STATUS_CHANGED);
	  signal_monitor(e->get_path(), e->get_path(), FILE_STATUS_CHANGED);
	  break;
	
	case FAMDeleted:
	  /*g_debug("FAM event: file '%s' deleted, req# %i\n",
		  e->get_path().c_str(),
		  FAMREQUEST_GETREQNUM(&event.fr));*/

	  e->set_change_news(FILE_STATUS_DELETED);
	  signal_monitor(e->get_path(), e->get_path(), FILE_STATUS_DELETED);

	  // Don't cancel monitoring, for example saving a file in most editors
	  // will appear as a deleted-created sequence.
	  //FAMCancelMonitor(&fam_, &event.fr);
	  break;

	case FAMCreated:
	  /*g_debug("FAM event: file '%s' created, req# %i\n",
		  e->get_path().c_str(),
		  FAMREQUEST_GETREQNUM(&event.fr));*/

	  e->set_change_news(FILE_STATUS_CREATED);
	  signal_monitor(e->get_path(), e->get_path(), FILE_STATUS_CREATED);
	  break;

	case FAMStartExecuting:
	case FAMStopExecuting:
	case FAMMoved:
	case FAMAcknowledge:
	case FAMExists:
	case FAMEndExist:
	  break;
	}
      }
    }
  }
  return true;
}

void
FileWatcher::setup_fam_for_entry(shared_ptr<Entry>& e)
{
  // We should handle fam events here to avoid possible deadlocks.
  on_fam_event(Glib::IO_IN);

  if (e->is_dir())
  {
    int res = FAMMonitorDirectory(&fam_,
				  e->get_path().c_str(),
				  &(e->get_request()),
				  e.get());
    if (res < 0)
      throw FamError("FAM error: can't monitor directory " + e->get_path());
    /*else
      g_debug("FAM req.# %i for %s",
	      FAMREQUEST_GETREQNUM(&(e->get_request())),
	      e->get_path().c_str()); // All ok.*/
  }
  else
  {
    int res = FAMMonitorFile(&fam_,
			     e->get_path().c_str(),
			     &(e->get_request()),
			     e.get());
    if (res < 0)
      throw FamError("FAM error: can't monitor file " + e->get_path());
    /*else
      g_debug("FAM req.# %i for %s",
	      FAMREQUEST_GETREQNUM(&(e->get_request())),
	      e->get_path().c_str());*/
  }
  
  // Again, handle any possible fam events.
  on_fam_event(Glib::IO_IN);
}

void
FileWatcher::add_file(const Glib::ustring& _path)
{
  Glib::ustring path(_path);

  //g_debug("Attempting to add %s to the watch list...", path.c_str());

  // Don't break down by going in /dev.
  if ((path.find("/dev/") != static_cast<Glib::ustring::size_type>(-1)) ||
      (path == "/dev"))
  {
    g_warning("Can't add a device file to the watch list.");
    return;
  }

  // Assure the path is in the right form.
  if ((! path.empty()) && (path[path.size() - 1] == '/'))
    path = path.substr(0, path.length() - 1);

  // Check if this entry already exists.
  EntryMap::iterator it = files_to_monitor_.find(path);
  if (it != files_to_monitor_.end())
  {
    //g_debug("Requested file entry is already being monitored.");
    return;
  }

  // Else, it's a proper new entry to be added.
  shared_ptr<Entry> new_entry(new Entry(path));

  if (Glib::file_test(path, Glib::FILE_TEST_EXISTS))
  {
    new_entry->set_is_dir(Glib::file_test(path, Glib::FILE_TEST_IS_DIR));
  }

  // If it doesn't exist it can also be monitored of course.
  // Doesn't matter if it's a dir but is being monitored as file.
  // We certainly can't determine this, and it's not necessary.
  setup_fam_for_entry(new_entry);
  files_to_monitor_[path] = new_entry;
  //g_debug("New file added to be monitored: %s", path.c_str());
}

void
FileWatcher::remove_file(const Glib::ustring& _path)
{
  // Find the corresponding entry, cancel monitoring, remove it from the map.

  Glib::ustring path(_path);
  //g_debug("Attempting to remove %s from the watch list...", path.c_str());

  // Assure the path is in the right form.
  if ((! path.empty()) && (path[path.size() - 1] == '/'))
    path = path.substr(0, path.length() - 1);

  Entry* req_entry = 0;
  EntryMap::iterator it(files_to_monitor_.begin());
  EntryMap::iterator end(files_to_monitor_.end());
	
  for ( ; it != end; ++it)
  {
    if (it->second->get_path() == path)
    {
      req_entry = it->second.get();
      FAMCancelMonitor(&fam_, &(req_entry->get_request()));
      files_to_monitor_.erase(it);
      break;
    }
  }
  //(req_entry) ? g_debug("ok") : g_debug("failed: entry not found");
}

} // namespace USBSink
