/*
 *  USBSink - flashdrivesdialog.cc
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "flashdrive.h"
#include "flashdrivesdialog.h"
#include "globalstrings.h"
#include "halclient.h"
#include "utils.h"

#include <algorithm>
#include <functional>
#include <iostream>
#include <sstream>
#include <string>
#include <list>

#include <glib/gi18n.h>
#include <glib/gmessages.h>

#include <sigc++/sigc++.h>

#include <glibmm/fileutils.h>
#include <glibmm/refptr.h>
#include <gtkmm.h>

#include <libglademm/xml.h>

using std::string;

namespace USBSink
{

// The magic const used for resizing the dialog after a click on the expander.
const unsigned short window_expansion = 133;

FlashDrivesDialog::FlashDrivesDialog(Gtk::Window& parent)
  :
  dialog_                (),
  treeview_flash_drives_ (0),
  button_mount_          (0),
  button_unmount_        (0),
  expander_properties_   (0),
  label_label_           (0),
  label_free_space_      (0),
  label_vendor_          (0),
  label_model_           (0),
  label_fs_              (0)
{
  load_xml();
  dialog_->set_transient_for(parent);
  dialog_->set_default_size(480, 270);
  connect_signals();
  fill_listview();
}

FlashDrivesDialog::~FlashDrivesDialog()
{}

void
FlashDrivesDialog::load_xml()
{
  using Gnome::Glade::Xml;

  try
  {
    const Glib::RefPtr<Xml> xml =
      Xml::create(glade_flashdrivesdialog_filename);
    
    Gtk::Dialog* dummy_dialog = 0;
    dialog_.reset(xml->get_widget("flashdrivesdialog", dummy_dialog));
    
    xml->get_widget("treeview_drives", treeview_flash_drives_);
    xml->get_widget("button_mount", button_mount_);
    xml->get_widget("button_unmount", button_unmount_);
    xml->get_widget("expander_properties", expander_properties_);
    xml->get_widget("label_label",   label_label_);
    xml->get_widget("label_free_space", label_free_space_);
    xml->get_widget("label_vendor", label_vendor_);
    xml->get_widget("label_model", label_model_);
    xml->get_widget("label_fs", label_fs_);
  }
  catch (const Gnome::Glade::XmlError& xml_err)
  {
    g_error("%s", xml_err.what().c_str());
  }
}

void
FlashDrivesDialog::connect_signals()
{
  // ui signals
  dialog_->signal_response().connect(
    sigc::mem_fun(*this, &FlashDrivesDialog::on_response));

  button_mount_->signal_clicked().connect(
    sigc::mem_fun(*this, &FlashDrivesDialog::on_mount_clicked));

  button_unmount_->signal_clicked().connect(
    sigc::mem_fun(*this, &FlashDrivesDialog::on_unmount_clicked));

  expander_properties_->property_expanded().signal_changed().connect(
    sigc::mem_fun(*this, &FlashDrivesDialog::on_properties_expanded));

  // halclient signals

  try
  {
    HalClient* hc = HalClient::instance();
    
    hc->signal_added.connect(
      sigc::mem_fun(*this, &FlashDrivesDialog::on_added));
    
    hc->signal_removed.connect(
      sigc::mem_fun(*this, &FlashDrivesDialog::on_removed));
    
    hc->signal_mounted.connect(
      sigc::mem_fun(*this, &FlashDrivesDialog::on_mounted));
    
    hc->signal_unmounted.connect(
      sigc::mem_fun(*this, &FlashDrivesDialog::on_unmounted));
  }
  catch (HalError& err)
  {
    g_warning("Hal exception occurred: %s", err.what().c_str());
  }
}

// Initializes the listview layout and appends any attached removables.
//
void
FlashDrivesDialog::fill_listview()
{
//  g_debug("FlashDrivesDialog::fill_listview()");

  // create the tree model
  listmodel_flash_drives_ = Gtk::ListStore::create(list_columns_);
  treeview_flash_drives_->set_model(listmodel_flash_drives_);

  // append columns
  treeview_flash_drives_->append_column(_("Label"), list_columns_.col_label);
  treeview_flash_drives_->append_column(_("Mount Point"),
					list_columns_.col_mnt_point);

  // talk to halclient
  try
  {
    HalClient* hc = HalClient::instance();
    drives_ = hc->get_drives();
    
    FlashDriveVector::const_iterator it(drives_->begin());
    FlashDriveVector::const_iterator end(drives_->end());
    
    if (it != end)
    {
      for (; it != end; ++it)
      {
	SharedFlashDrive sfd = *it;
//        g_debug("displaying information about %s on %s",
//                (*it)->get_label().c_str(), (*it)->get_udi().c_str());

	// append a new row
	Gtk::TreeModel::Row row = *(listmodel_flash_drives_->append());
	row[list_columns_.col_udi] = sfd->get_udi();
	row[list_columns_.col_label] = sfd->get_label();
	(sfd->is_mounted())
	  ? row[list_columns_.col_mnt_point] = sfd->get_mount_point()
	  : row[list_columns_.col_mnt_point] = _("(not mounted)");

	update_expander(sfd);
      }
    }
    else
    {
      // no drives detected
      button_mount_->set_sensitive(false);
      button_unmount_->set_sensitive(false);
      expander_properties_->set_sensitive(false);
    }
    
    list_selection_ = treeview_flash_drives_->get_selection();
    // connect to the item-selection signal
    list_selection_->signal_changed().connect(
      sigc::mem_fun(*this, &FlashDrivesDialog::on_drive_selection_changed));
  }
  catch (HalError& err)
  {
    g_warning("HAL exception: %s", err.what().c_str());
  }
}

void
FlashDrivesDialog::update_expander(const SharedFlashDrive& sfd)
{
//  g_debug("FlashDrivesDialog::update_expander()");

  label_label_->set_text(sfd->get_label());
  label_vendor_->set_text(sfd->get_vendor());
  label_model_->set_text(sfd->get_model());
  
  if (sfd->is_mounted())
  {
    using std::list;
    using std::string;
    using std::ostringstream;

//    g_debug("volume is mounted");
    
    ostringstream oss;
    
//    g_debug("displaying space info");
    double free_space = sfd->get_free_space() / 1024 / 1024;
    double total_space = sfd->get_size() / 1024 / 1024;
    
    // free space out of total space
    oss << Util::i18n_printf(_("%u out of %u MB"),
                             static_cast<guint>(free_space),
                             static_cast<guint>(total_space));
    label_free_space_->set_text(oss.str());
    
    label_fs_->set_text(sfd->get_fs_type());
  }
  else
  {
//    g_debug("volume is not mounted");

    const Glib::ustring na(_("N/A"));
    
    label_free_space_->set_text(na);
    label_fs_->set_text(na);
  }
}

void
FlashDrivesDialog::on_drive_selection_changed()
{
  Gtk::TreeModel::iterator it = list_selection_->get_selected();

  if (it)
  {
    Gtk::TreeModel::Row row = *it;
    string udi = row[list_columns_.col_udi];
    
    // find the corresponding FlashDrive and update the labels in expander
    
    FlashDriveVector::const_iterator fit(
      find(drives_->begin(), drives_->end(), udi));
    
    if (fit != drives_->end())
    {
      if ((*fit)->is_mounted())
      {
	button_mount_->set_sensitive(false);
	button_unmount_->set_sensitive(true);
      }
      else
      {
	button_mount_->set_sensitive(true);
	button_unmount_->set_sensitive(false);
      }
      update_expander(*fit);
    }
    expander_properties_->set_sensitive(true);
  }
  else // nothing is selected - the list is empty
  {
    button_mount_->set_sensitive(false);
    button_unmount_->set_sensitive(false);
    expander_properties_->set_sensitive(false);
  }
}

void
FlashDrivesDialog::on_mount_clicked()
{
  Gtk::TreeModel::Row row;
  
  if (row = *(list_selection_->get_selected()))
  {
    string udi = row[list_columns_.col_udi];
    
    FlashDriveVector::const_iterator it(
      find(drives_->begin(), drives_->end(), udi));
    
    if (it != drives_->end())
    {
      try
      {
	SharedFlashDrive flash(*it);
	flash->mount();
	button_mount_->set_sensitive(false);
	button_unmount_->set_sensitive(true);
	update_expander(flash);
      }
      catch (MountError& err)
      {
	Glib::ustring message(_("Cannot mount volume."));
	Gtk::MessageDialog err_dialog(*(dialog_->get_transient_for()),
				      message, false, Gtk::MESSAGE_ERROR,
				      Gtk::BUTTONS_OK, true);
	Glib::ustring err_details = err.what();
	err_dialog.set_secondary_text(err_details);
	err_dialog.run();
      }
    }
  }
}

void
FlashDrivesDialog::on_unmount_clicked()
{
  Gtk::TreeModel::Row row;
  if (row = *(list_selection_->get_selected()))
  {
    string udi = row[list_columns_.col_udi];
    
    FlashDriveVector::const_iterator it(
      find(drives_->begin(), drives_->end(), udi));
    
    if (it != drives_->end())
    {
      try
      {
	SharedFlashDrive flash(*it);
	flash->umount();
	button_mount_->set_sensitive(true);
	button_unmount_->set_sensitive(false);
	update_expander(flash);
      }
      catch (MountError& err)
      {
	Glib::ustring message(_("Cannot unmount volume."));
	Gtk::MessageDialog err_dialog(*(dialog_->get_transient_for()),
				      message, false, Gtk::MESSAGE_ERROR,
				      Gtk::BUTTONS_OK, true);
	Glib::ustring err_details = err.what();
	err_dialog.set_secondary_text(err_details);
	err_dialog.run();
      }
    }
  }
}

// On a click on the expander, grows/shrinks the dialog and updates
// expander's label.
//
void
FlashDrivesDialog::on_properties_expanded()
{
  int width, height;
  
  dialog_->get_size(width, height);
  
  if (expander_properties_->get_expanded())
  {
    dialog_->resize(width, height + window_expansion);
    expander_properties_->set_label(_("Hide Device Properties"));
  }
  else
  {
    dialog_->resize(width, height - window_expansion);
    expander_properties_->set_label(_("Show Device Properties"));
  }
}

void
FlashDrivesDialog::on_added(const SharedFlashDrive& flash)
{
  HalClient* hc = HalClient::instance();
  drives_ = hc->get_drives(); // stay up2date
  
  Gtk::TreeModel::Row row = *(listmodel_flash_drives_->append());
  
  row[list_columns_.col_udi] = flash->get_udi();
  row[list_columns_.col_label] = flash->get_label();
  row[list_columns_.col_mnt_point] = _("(not mounted)");
  
  // no need to update the expander, as there is no
  // "automatic selection of freshly added drives"
}

void
FlashDrivesDialog::on_removed(const SharedFlashDrive& flash)
{
  HalClient* hc = HalClient::instance();
  drives_ = hc->get_drives();
  
  tree_children children = listmodel_flash_drives_->children();
  tree_children::iterator iter = children.begin();
  tree_children::iterator end = children.end();
  string flash_udi = flash->get_udi();

  // TODO: consider a functor and find_if here as well
  for (; iter != end; ++iter)
  {
    string row_udi = (*iter)[list_columns_.col_udi];
    if (row_udi == flash_udi)
    {
      break;
    }
  }
  if (iter) // use implemented operator bool()
  {
    listmodel_flash_drives_->erase(*iter);
  }
  
  // If the last flash was removed, shrink the expander.
  
  if ((listmodel_flash_drives_->children().begin()
       == listmodel_flash_drives_->children().end())
      && expander_properties_->get_expanded())
  {
    expander_properties_->set_expanded(false);
  }
}

void
FlashDrivesDialog::on_mounted(const SharedFlashDrive& flash)
{
  Gtk::TreeModel::Row row;
  
  if (row = *(list_selection_->get_selected()))
  {
    string udi = row[list_columns_.col_udi];
    
    if (flash->get_udi() == udi)
    {
      button_mount_->set_sensitive(false);
      button_unmount_->set_sensitive(true);
      update_expander(flash);
    }
  }

  // Even if the flash in question is currently not selected,
  // its' mount point column needs to be updated.
  // It's possible that at first there were no flash drives,
  // and then this one was added - in that case it will appear
  // in the treeview, but *nothing will be selected*, and we
  // must update it's info of course. That's why this is
  // outside if_(anything_)selected.
	
  tree_children children = listmodel_flash_drives_->children();
  tree_children::iterator iter = children.begin();
  tree_children::iterator end = children.end();
  string flash_udi = flash->get_udi();
  
  // TODO: this stl way
  for (; iter != end; ++iter)
  {
    string row_udi = (*iter)[list_columns_.col_udi];
    if (row_udi == flash_udi)
    {
      break;
    }
  }
  if (iter) // just in case
  {
    Gtk::TreeModel::Row row = *iter;
    row[list_columns_.col_mnt_point] = flash->get_mount_point();
  }
}

void
FlashDrivesDialog::on_unmounted(const SharedFlashDrive& flash)
{
  Gtk::TreeModel::Row row;
  
  if (row = *(list_selection_->get_selected()))
  {
    string udi = row[list_columns_.col_udi];
    
    if (flash->get_udi() == udi)
    {
      button_mount_->set_sensitive(true);
      button_unmount_->set_sensitive(false);
      update_expander(flash);
    }
    
    tree_children children = listmodel_flash_drives_->children();
    tree_children::iterator iter = children.begin();
    tree_children::iterator end = children.end();

    // TODO: this stl way
    for (; iter != end; ++iter)
    {
      if ((*iter)[list_columns_.col_label] == flash->get_label())
      {
	break;
      }
    }
    if (iter)
    {
      Gtk::TreeModel::Row row = *iter;
      row[list_columns_.col_mnt_point] = _("Not mounted");
    }
  }
}

} // namespace USBSink
