/*
 *  USBSink - halclient.cc
 *
 *  Copyright (C) 2005 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "flashdrive.h"
#include "halclient.h"

#include <string.h>

#include <algorithm>
#include <memory>
#include <string>
#include <vector>

#include <tr1/memory>

#include <glib/gmessages.h>

#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>
#include <libhal-storage.h>

#include <sigc++/sigc++.h>

using std::auto_ptr;
using std::string;
using std::vector;

using std::tr1::shared_ptr;

namespace
{

// Libhal callbacks.

// When this function is called, the volume is not mounted by a volume
// manager yet, but it needs to be available to the user.
// Likewise, on_removed executes before unmount.
// (Un)mounting triggers on_hal_property_modified.
static void
on_hal_device_added(LibHalContext* ctx, const char* udi)
{
//  g_debug("on_hal_device_added(): %s", udi);
  
  LibHalVolume* volume;
  if ((volume = libhal_volume_from_udi(ctx, udi)))
  {
    LibHalDrive* drive =
      libhal_drive_from_udi(ctx, libhal_volume_get_storage_device_udi(volume));
    
    if (libhal_drive_get_bus(drive) == LIBHAL_DRIVE_BUS_USB)
    {
//      g_debug("%s is a volume on USB", udi);

      USBSink::HalClient* hc = USBSink::HalClient::instance();
      hc->add_drive(drive, volume);
    }
    libhal_drive_free(drive);
  }
  libhal_volume_free(volume);
}

// There are two ways for a volume to become unavailable:
// when the user unplugs a drive with a mounted volume and when she
// unmounts the volume, but the drive stays plugged in.

// This function is executed when a drive is unplugged from the computer,
// and it is then when we need to remove it from the drives_ vector.
// Unmounting a volume should stop all related tasks, but we need to keep
// watching it.
static void
on_hal_device_removed(LibHalContext* /* ctx */, const char* _udi)
{
//  g_debug("on_hal_device_removed(): %s", _udi);

  string udi(_udi);

  USBSink::HalClient* hc = USBSink::HalClient::instance();

  USBSink::SharedFlashDrive flash = hc->find_flash_by_udi(udi);

  if (flash.get())
  {
//    g_debug("removing flash %s", flash->get_label().c_str());
    hc->remove_drive(flash);
  }
  else
  {
//    g_debug("could not get a pointer to a FlashDrive for %s", _udi);
  }
}

// When a drive is plugged or unplugged, first changes the volume.mount_point
// and then volume.is_mounted property. If there is a volume manager, that is.
// For tasks and gui, what matters are the mount/unmount events, so appropriate
// signals are emitted here.
static void
on_hal_property_modified(LibHalContext* ctx,
			 const char* _udi,
			 const char* key,
			 dbus_bool_t /* is_removed */,
			 dbus_bool_t /* is_added */)
{
//  g_debug("on_hal_property_modified(): something's changed on %s: %s",
//          _udi, key);

  if (strcmp(key, "volume.is_mounted") == 0)
  {
    USBSink::HalClient* hc = USBSink::HalClient::instance();
    string udi(_udi);
    USBSink::SharedFlashDrive flash = hc->find_flash_by_udi(udi);

    if (flash)
    {
//      g_debug("getting volume from udi");
      LibHalVolume* volume = libhal_volume_from_udi(ctx, _udi);


//      g_debug("getting storage device + drive");
      LibHalDrive* drive =
        libhal_drive_from_udi(ctx,
                              libhal_volume_get_storage_device_udi(volume));

//      g_debug("emitting signals");
    
      if (libhal_volume_is_mounted(volume))
      {
        flash->load_dev_info(drive, volume);
        hc->emit_signal_on_flash(flash, "mounted");
      }
      else
      {
        flash->load_dev_info(drive, volume); // will set those "N/A" values
        hc->emit_signal_on_flash(flash, "unmounted");
      }
    
      libhal_drive_free(drive);
      libhal_volume_free(volume);
    }
    else
    {
//      g_debug("Unexpected: could not get a pointer to FlashDrive for %s",
//              _udi);
    }
  }
}

} // anonymous namespace

namespace USBSink
{

auto_ptr<HalClient> HalClient::instance_;

HalClient*
HalClient::instance()
{
  if (! instance_.get())
  {
    instance_.reset(new HalClient);
  }
  return instance_.get();
}

HalClient::HalClient()
{
  if (! init_hal())
  {
    g_warning("Can't connect to HAL! Please start hald.\n");
    throw HalError("Error initializing HAL.");
  }
}

HalClient::~HalClient()
{
  disconnect_from_hal();
  drives_.clear();
}

bool
HalClient::init_hal()
{
//  g_debug("HalClient::init_hal()");

  if (! (hal_ctx_ = libhal_ctx_new()))
  {
    g_warning("Failed to create new HAL context!");
    return false;
  }
  
  DBusError error;
  dbus_error_init(&error);

  if (! integrate_hal_into_mainloop(&error))
  {
    g_warning("Failed to integrate HAL into mainloop: %s", error.message);
    dbus_error_free(&error);
    return false;
  }

  set_hal_callbacks();
  
  if (! libhal_device_property_watch_all(hal_ctx_, &error))
  {
    g_warning("Failed to watch all HAL properties: %s", error.message);
    dbus_error_free(&error);
    return false;
  }
  
  if (! libhal_ctx_init(hal_ctx_, &error))
  {
    g_warning("Failed to initialize HAL: %s", error.message);
    dbus_error_free(&error);
    return false;
  }

  if (! get_drives_at_startup(&error))
  {
    g_warning("Failed to get info about currently available devices: %s",
	      error.message);
    dbus_error_free(&error);
    return false;
  }
	
  return true;
}

bool
HalClient::integrate_hal_into_mainloop(DBusError* error)
{
  dbus_conn_ = dbus_bus_get(DBUS_BUS_SYSTEM, error);
  
  if (dbus_error_is_set(error))
  {
    return false;
  }
  
  dbus_connection_setup_with_g_main(dbus_conn_, NULL);
  
  libhal_ctx_set_dbus_connection(hal_ctx_, dbus_conn_);
  
  return true;
}

void
HalClient::set_hal_callbacks()
{
  libhal_ctx_set_device_added(hal_ctx_, on_hal_device_added);
  libhal_ctx_set_device_removed (hal_ctx_, on_hal_device_removed);
  libhal_ctx_set_device_property_modified(hal_ctx_, on_hal_property_modified);
}

// Should be called only in the ctor.
bool
HalClient::get_drives_at_startup(DBusError* error)
{
  char** devices;
  int num_devices;

//  g_debug("HalClient::get_drives_at_startup()");
  
  devices = libhal_get_all_devices(hal_ctx_, &num_devices, error);
  if (dbus_error_is_set(error))
  {
    return false;
  }
  else
  {
    for (int i = 0; i < num_devices; ++i)
    {
      LibHalDrive* drive;
      if ((drive = libhal_drive_from_udi(hal_ctx_, devices[i])))
      {
	// The device is a drive. Now...
        // (don't do (libhal_drive_get_type(drive) == LIBHAL_DRIVE_TYPE_REMOVABLE_DISK)
        // external HD is not a removable, for instance, and this is easier)
	if (libhal_drive_get_bus(drive) == LIBHAL_DRIVE_BUS_USB)
	{
	  char** volumes;
	  int num_volumes;
	  
	  if ((volumes = libhal_drive_find_all_volumes(hal_ctx_,
						       drive,
						       &num_volumes))
	      && (num_volumes))
	  {
//            g_debug("%i volumes on drive %s", num_volumes, devices[i]);

	    // No need for a loop, each flash drive has only one volume.
            for (int vol_index = 0; vol_index < num_volumes; ++vol_index)
            {
              LibHalVolume* volume
                = libhal_volume_from_udi(hal_ctx_, volumes[vol_index]);
	    
              SharedFlashDrive flash(
                new FlashDrive(dbus_conn_, drive, volume));
              drives_.push_back(flash);
	    
              libhal_volume_free(volume);
            }

            libhal_free_string_array(volumes);
	  }
          else
          {
//            g_debug("Could not determine volumes on drive %s", devices[i]);
          }
	}
      }
      libhal_drive_free(drive);
    }
  }
  libhal_free_string_array(devices);
  
  return true;
}

void
HalClient::disconnect_from_hal()
{
  libhal_ctx_shutdown(hal_ctx_, NULL);
  libhal_ctx_free(hal_ctx_);
}

void
HalClient::add_drive(LibHalDrive* drive, LibHalVolume* volume)
{
//  g_debug("HalClient::add_drive(): entered");

  SharedFlashDrive flash(new FlashDrive(dbus_conn_, drive, volume));
  drives_.push_back(flash);
  signal_added.emit(flash);
}

void
HalClient::remove_drive(SharedFlashDrive flash)
{
//  g_debug("HalClient::remove_drive(): removing %s", flash->get_udi().c_str());

  signal_removed.emit(flash);
  drives_.erase(remove(drives_.begin(), drives_.end(), flash),
		drives_.end());
}

SharedFlashDrive
HalClient::find_flash_by_udi(string& udi)
{
  FlashDrive* fd = 0;
  SharedFlashDrive sfd(fd);
  FlashDriveVector::iterator it = drives_.begin();
  
  for (; it != drives_.end(); ++it)
  {
    if ((*it)->get_udi() == udi)
    {
      sfd = *it;
      break;
    }
  }
  return sfd;
}

void
HalClient::emit_signal_on_flash(SharedFlashDrive sfd, const char* signal_name)
{
  if (strcmp(signal_name, "mounted") == 0)
  {
    signal_mounted.emit(sfd);
  }
  else if (strcmp(signal_name, "unmounted") == 0)
  {
    signal_unmounted.emit(sfd);
  }
}

} // namespace USBSink
