/*
 *  USBSink - logviewdialog.cc
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "globalstrings.h"
#include "logviewdialog.h"
#include "task.h"
#include "taskmanager.h"

#include <list>
#include <tr1/memory>

#include <glib/gi18n.h>

#include <gtk/gtk.h>
#include <gtkmm.h>

#include <libglademm/xml.h>

using std::list;

namespace
{

static void
save_textbuffer_to_file(Gtk::Window& parent,
                        const Glib::RefPtr<Gtk::TextBuffer>& buffer,
                        const Glib::ustring& filename)
{
  const char* text = buffer->get_text().c_str();
  GError* error = NULL;

  if (g_file_set_contents (filename.c_str(), text, -1, &error))
  {
    // TODO: update main status bar
  }
  else
  {
    Gtk::MessageDialog err_dialog(parent,
				  _("Could not save task log"),
				  false, Gtk::MESSAGE_ERROR,
				  Gtk::BUTTONS_OK, true);
    
    Glib::ustring err_details(error->message);
    err_dialog.set_secondary_text(err_details);
    err_dialog.run();

    g_error_free (error);
  }
}

} // anonymous namespace

namespace USBSink
{

typedef std::tr1::shared_ptr<TaskHandling::Task> SharedTask;

LogViewDialog::LogViewDialog(Gtk::Window& parent)
  :
  treeview_task_list_ (0),
  hpaned_      (0),
  textview_log_(0),
  button_clear_(0),
  button_save_ (0),
  button_ok_   (0),
  selected_task_id_(0)
{
  load_xml();
  dialog_->set_transient_for(parent);
  dialog_->set_default_size(600, 300);
  setup_task_list();
  connect_signals();
}

LogViewDialog::~LogViewDialog()
{
}

void
LogViewDialog::load_xml()
{
  using Gnome::Glade::Xml;

  try
  {
    const Glib::RefPtr<Xml> xml = Xml::create(glade_logviewdialog_filename);

    Gtk::Dialog* dialog_ptr = 0;
    dialog_.reset(xml->get_widget("logviewdialog", dialog_ptr));

    xml->get_widget("treeview_task_list", treeview_task_list_);
    xml->get_widget("textview_log", textview_log_);

    xml->get_widget("button_clear", button_clear_);
    xml->get_widget("button_save", button_save_);
    xml->get_widget("button_ok", button_ok_);
  }
  catch (const Gnome::Glade::XmlError& xml_err)
  {
    g_error("%s", xml_err.what().c_str());
  }
}

void
LogViewDialog::connect_signals()
{
  button_clear_->signal_clicked().connect(
    sigc::mem_fun(*this, &LogViewDialog::on_button_clear_clicked));

  button_save_->signal_clicked().connect(
    sigc::mem_fun(*this, &LogViewDialog::on_button_save_clicked));

  button_ok_->signal_clicked().connect(
    sigc::mem_fun(*this, &LogViewDialog::on_button_ok_clicked));
}

void
LogViewDialog::setup_task_list()
{
  // create the tree model
  treemodel_tasks_ = Gtk::ListStore::create(columns_tasks_);
  treeview_task_list_->set_model(treemodel_tasks_);

  // set up the treeview selection object
  task_list_selection_ = treeview_task_list_->get_selection();
  task_list_selection_->set_mode(Gtk::SELECTION_SINGLE);

  // append column headers
  treeview_task_list_->append_column(_("Task"), columns_tasks_.col_name);

  // load the task names from TaskManager
  TaskHandling::Manager* manager = TaskHandling::Manager::instance();
  list<SharedTask> tasks = manager->get_tasks();

  // get them in the treeview
  list<SharedTask>::iterator it(tasks.begin());
  list<SharedTask>::iterator end(tasks.end());

  for ( ; it != end; ++it )
  {
    // add a new row to the treeview
    Gtk::TreeModel::Row row = *(treemodel_tasks_->append());

    row[columns_tasks_.col_id] = (*it)->get_id();
    row[columns_tasks_.col_name] = (*it)->get_name();
  }

  task_list_selection_->set_select_function(
    sigc::mem_fun(*this, &LogViewDialog::on_treeview_select_function));
}

bool
LogViewDialog::on_treeview_select_function(
                  const Glib::RefPtr<Gtk::TreeModel>& /* model */,
                  const Gtk::TreeModel::Path& path,
                  bool path_currently_selected)
{
  Gtk::TreeModel::Row row = *(treemodel_tasks_->get_iter(path));

  if ((row[columns_tasks_.col_id] != selected_task_id_) &&
      (! path_currently_selected)) // this was just empirically determined...
  {
    // selection changed
    selected_task_id_ = row[columns_tasks_.col_id];

    // change text buffer to display
    TaskHandling::Manager* manager = TaskHandling::Manager::instance();
    SharedTask task = manager->get_task(selected_task_id_);
    textview_log_->set_buffer(task->get_logger()->get_text_buffer());
  }

  return true;
}

void
LogViewDialog::on_button_clear_clicked()
{
  TaskHandling::Manager* manager = TaskHandling::Manager::instance();
  Glib::RefPtr<Gtk::TextBuffer> buffer =
    manager->get_task(selected_task_id_)->get_logger()->get_text_buffer();
  buffer->erase(buffer->begin(), buffer->end());
}

void
LogViewDialog::on_button_save_clicked()
{
  GtkWidget* dialog;
  
  dialog = gtk_file_chooser_dialog_new (_("Save selected task log to file"),
					GTK_WINDOW(dialog_->gobj()),
					GTK_FILE_CHOOSER_ACTION_SAVE,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
					NULL);
  gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog),
						  TRUE);

  SharedTask task(
    TaskHandling::Manager::instance()->get_task(selected_task_id_));
  Glib::ustring name(task->get_name());
  name += ".log";

  gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), name.c_str());

  if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
  {
    Glib::ustring filename(
      gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog)));
    save_textbuffer_to_file(*(dialog_->get_transient_for()),
			    task->get_logger()->get_text_buffer(),
			    filename);
  }

  gtk_widget_destroy (dialog);
}

void
LogViewDialog::on_button_ok_clicked()
{
  dialog_->hide();
}

} // namespace USBSink
