// -*- c++ -*-
/*
 *  USBSink - newtaskdialog.h
 *
 *  Copyright (C) 2005-2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _NEWTASKDIALOG_H_
#define _NEWTASKDIALOG_H_

#include <list>
#include <memory>
#include <string>

#include <tr1/memory>

#include <gtkmm.h>

#include <sigc++/sigc++.h>

namespace Gnome { namespace Vfs { class Uri; } }

namespace USBSink
{

const int ms_in_min = 60000;

class AddFilesDialog;
class FlashDrive;
namespace TaskHandling { class Task; }

typedef std::tr1::shared_ptr<FlashDrive> SharedFlashDrive;

enum IgnoreFilesDialogResponse
{
  USBSINK_RESPONSE_CREATE,
  USBSINK_RESPONSE_GO_BACK,
  USBSINK_RESPONSE_CONTINUE
};

class NewTaskDialog : public sigc::trackable
{
public:

  explicit NewTaskDialog(Gtk::Window& parent);
  virtual ~NewTaskDialog();

  Gtk::Dialog* get_dialog() { return dialog_.get(); }

  /// eg void on_file_list_editor_request(const Glib::ustring& list_name);
  sigc::signal<void, const Glib::ustring&> signal_file_list_editor_request;

protected:

  // tree model column classes
  //
  class ComboModelColumns : public Gtk::TreeModel::ColumnRecord
  {
  public:

    ComboModelColumns()
    { add(col_udi); add(col_label); }
    
    // col_udi is not displayed, but it is used to identify selected drive,
    // in case there are more than one with the same display name.
    //
    Gtk::TreeModelColumn<std::string> col_udi;
    Gtk::TreeModelColumn<Glib::ustring> col_label;
  };

  class FileListModelColumns : public Gtk::TreeModel::ColumnRecord
  {
  public:

    FileListModelColumns()
      {
	add(col_name);
	add(col_modified);
	add(col_mime_type);
	add(col_full_path);
      }

    Gtk::TreeModelColumn<Glib::ustring> col_name;
    Gtk::TreeModelColumn<Glib::ustring> col_modified;
    Gtk::TreeModelColumn<Glib::ustring> col_mime_type;
    Gtk::TreeModelColumn<Glib::ustring> col_full_path;
  };

  // functor to populate the combobox with connected removable drives
  //
  struct AddToComboBox
    : public std::binary_function<SharedFlashDrive, NewTaskDialog*, void>
  {
    void operator()(const SharedFlashDrive flash,
                    const NewTaskDialog* dlg) const;
  };

  // signal handlers
  //
  virtual void on_file_list_editor_clicked();

  virtual void on_toolbutton_add();
  virtual void on_toolbutton_remove();
  virtual void on_toolbutton_clear();
  virtual void on_toolbutton_import();

  virtual void on_browse_dir();

  virtual void on_response(int response_id);

  virtual bool file_is_in_list(const Glib::ustring& uri);

  virtual void on_addfilesdialog_hide();
  virtual void on_addfilesdialog_files_selected(
    std::list< Glib::RefPtr<Gnome::Vfs::Uri> >& uris);
  
  virtual void on_treeview_drag_data_received(
    const Glib::RefPtr<Gdk::DragContext>& context,
    int x, int y,
    const Gtk::SelectionData& selection_data,
    guint info, guint time);

  // member functions
  //
  virtual void load_xml();
  virtual void connect_signals();

  void populate_drive_combobox();
  void setup_file_list_layout();

  virtual void add_uri_to_treeview(const Glib::RefPtr< Gnome::Vfs::Uri >& uri);
  virtual std::list< Glib::RefPtr<Gnome::Vfs::Uri> > get_source_files();

  void check_is_task_name_duplicate();
  void handle_duplicate_task_name(const Glib::ustring& name);
  void register_task();

  // Called on_ok.
  // Throws InputError.
  //
  virtual void check_input();

  std::tr1::shared_ptr<TaskHandling::Task> create_task_from_input();

  // dialog ptrs
  //
  std::auto_ptr<Gtk::Dialog> dialog_;
  std::auto_ptr<AddFilesDialog> addfiles_dialog_;

  // child widgets
  //
  Gtk::Entry*                  entry_task_name_;
  Gtk::ComboBox*               combobox_target_drive_;
  ComboModelColumns            columns_removables_;
  Glib::RefPtr<Gtk::ListStore> treemodel_drives_;
  Gtk::ComboBox*               combobox_data_flow_;
  Gtk::ComboBox*               combobox_file_list_type_;
  Gtk::TextView*               textview_file_list_;
  Gtk::Button*                 button_file_list_editor_;
  Gtk::ComboBox*               combobox_start_;
  Gtk::SpinButton*             spinbutton_timeout_;

  Gtk::Toolbar*                    toolbar_;
  Gtk::ToolButton*                 toolbutton_add_;
  Gtk::ToolButton*                 toolbutton_remove_;
  Gtk::ToolButton*                 toolbutton_clear_;
  Gtk::ToolButton*                 toolbutton_import_;
  Gtk::TreeView*                   treeview_file_list_;
  FileListModelColumns             columns_files_;
  Glib::RefPtr<Gtk::ListStore>     treemodel_files_;
  Glib::RefPtr<Gtk::TreeSelection> file_list_selection_;

  Gtk::Entry*       entry_target_dir_;
  Gtk::Button*      button_browse_dir_;

  Glib::RefPtr<Gtk::TextBuffer> textview_buffer_;

  // drag-and-drop targets
  //
  typedef std::list<Gtk::TargetEntry> list_dnd_targets;
  list_dnd_targets dnd_targets_;
};

// exception types
//
class InputError
{
 public:
  InputError(const Glib::ustring& msg) : msg_(msg) {}
  Glib::ustring what() const { return msg_; }
 private:
  Glib::ustring msg_;
};

class IgnoreFilesException
{
 public:
  IgnoreFilesException(const int response_id) : response_(response_id) {}
  int what() const { return response_; }
 private:
  int response_;
};

} // namespace USBSink

#endif // _NEWTASKDIALOG_H_
