/*
 *  USBSink - utils.cc
 *
 *  Copyright (C) 2006 Marko Anastasov
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "globalstrings.h"
#include "utils.h"

#include <errno.h>
#include <time.h>
#include <sys/time.h>

#include <fstream>
#include <iostream>
#include <string>

#include <glib/gstdio.h>
#include <glibmm.h>

#include <gtkmm/textbuffer.h>

#include <pcrecpp.h>

using std::ifstream;
using std::ofstream;
using std::list;
using std::string;

namespace USBSink
{

namespace Util
{

Glib::ustring
i18n_printf(const Glib::ustring& format, ...)
{
  va_list args;
  va_start(args, format);
  gchar* cstr = g_strdup_vprintf(format.c_str(), args);
  Glib::ustring str(cstr);
  g_free(cstr);
  
  return str;
}

namespace Config
{

string
get_rc_dir()
{
  return (Glib::get_home_dir() + G_DIR_SEPARATOR_S + USBSink::rc_dir +
	  G_DIR_SEPARATOR + "usbsink");
}

string
get_task_dir()
{
  return (get_rc_dir() + G_DIR_SEPARATOR_S + USBSink::task_dir);
}

string
get_file_list_dir()
{
  return (get_rc_dir() + G_DIR_SEPARATOR_S + USBSink::file_list_dir);
}

void
create_common_ignore_file_list()
{
  string filename = get_file_list_dir() + G_DIR_SEPARATOR_S + "common";

  if (! Glib::file_test(filename, Glib::FILE_TEST_EXISTS))
  {
    ofstream os(filename.c_str());

    if (! os)
    {
      g_warning("Unable to open file: %s\n", filename.c_str());
      return;
    }

    os << "(\\.).*" << std::endl;
    os << ".*~" << std::endl;
    os << ".*\\.swp" << std::endl;
    os << ".*\\.bak" << std::endl;
  }
}

Glib::ustring
get_file_list_format_string()
{
  Glib::ustring format(file_list_format_left_terminator);
  format += "%s";
  format += file_list_format_right_terminator;

  return format;
}

} // namespace Config

namespace File
{

int
count_files_in_dir(const string& path, bool recursive)
{
  if (! (Glib::file_test(path, Glib::FILE_TEST_IS_DIR) ||
	 (Glib::file_test(path, Glib::FILE_TEST_EXISTS))))
  {
    return 0;
  }

  unsigned count = 0;

  try
  {
    Glib::Dir dir(path);
    list<string> items(dir.begin(), dir.end());
		
    for (list<string>::iterator it = items.begin(); it != items.end(); ++it)
    {
      string item = Glib::build_filename(path, *it);
			
      if (Glib::file_test(item, Glib::FILE_TEST_IS_DIR) && recursive)
      {
	count += count_files_in_dir(item, true);
      }
      else if (Glib::file_test(item, Glib::FILE_TEST_IS_REGULAR |
			       Glib::FILE_TEST_IS_SYMLINK |
			       Glib::FILE_TEST_IS_EXECUTABLE))
      {
	count++;
      }
    }
  }
  catch (Glib::FileError& ex)
  {
    g_warning("Cannot open directory %s: %s",
	      path.c_str(),
	      Glib::strerror(ex.code()).c_str());
  }
  
  return count;
}

/**
 * Eg for file /home/marko/some/cool/path/to/file
 *              |___|_____|____|____|____|__|
 *
 * All directories which do not exist will be created.
 * A convenience wrapper around g_mkdir_with_parents().
 *              
 */
void
make_missing_dirs_for_file(const Glib::ustring& _path)
{
  if (! Glib::path_is_absolute(_path))
    return;

  Glib::ustring path(Glib::path_get_dirname(_path));

  if (! Glib::file_test(path, Glib::FILE_TEST_EXISTS))
    if (g_mkdir_with_parents(path.c_str(), S_IRWXU) == -1)
    {
      g_warning("Could not mkdir %s with parents: %s",
		path.c_str(),
		g_strerror(errno));
    }
}

/**
 * @return A list of files inside a directory. Non-recursive.
 * @throw Glib::FileError
 */
list<std::string>
get_files_in_dir(const std::string& path)
{
  Glib::Dir root_dir(path);
  list<std::string> files(root_dir.begin(), root_dir.end());
  
  list<std::string>::iterator it(files.begin());
  list<std::string>::iterator end(files.end());

  for ( ; it != end; ++it)
    *it = Glib::build_filename(path, *it);

  return files;
}

/**
 * @return First line in a file.
 * @throw Glib::FileError
 */
Glib::ustring
peek_first_line(const std::string& path)
{
  Glib::RefPtr<Glib::IOChannel> channel =
    Glib::IOChannel::create_from_file(path, "r");

  Glib::ustring line_buf;

  channel->read_line(line_buf);
  channel->close();

  return String::trim_whitespace(line_buf);
}

} // namespace File

namespace Regexp
{

bool
validate(Glib::ustring& str)
{
  bool result = false;

  pcrecpp::RE regexp(str.c_str(), pcrecpp::UTF8());

  result = regexp.error().empty();

  return result;
}

} // namespace Regexp

namespace String
{

list<Glib::ustring>
load_file_to_list(const Glib::ustring& filename)
{
  ifstream is(filename.c_str());
  list<Glib::ustring> contents;

  if (! is)
  {
    g_warning("Unable to open file: %s\n", filename.c_str());
    return contents;
  }

  Glib::ustring buf;

  while (is >> buf)
  {
    contents.push_back(buf);
  }

  return contents;  
}

/**
 * Returns a list of trimmed lines from a Gtk::TextBuffer.
 */
std::list<Glib::ustring>
get_lines_from_text_buffer(const Glib::RefPtr<Gtk::TextBuffer>& buffer)
{
  std::list<Glib::ustring> lines;
  Glib::ustring buf;

  Gtk::TextIter line_begin(buffer->begin());
  Gtk::TextIter line_end(buffer->begin());

  while (line_end.forward_to_line_end())
  {
    buf = buffer->get_slice(line_begin, line_end, false);
    buf = trim_whitespace(buf);
    lines.push_back(buf);

    line_begin.forward_line();
    line_end.forward_line();
  }

  // last line  
  buf = buffer->get_slice(line_begin, line_end, false).raw();
  buf = trim_whitespace(buf);

  if (! buf.empty())
  {
    lines.push_back(buf);
  }
  
  return lines;
}

Glib::ustring
trim_whitespace(const Glib::ustring& text)
{
  Glib::ustring::const_iterator it(text.begin());
  Glib::ustring::const_iterator end(text.end());

  while(it != end && Glib::Unicode::isspace(*it))
    ++it;

  Glib::ustring::const_iterator temp(end);

  while(it != temp && Glib::Unicode::isspace(*--temp))
    end = temp;

  return Glib::ustring(it, end);
}

int
count_char(const Glib::ustring& text, gunichar c)
{
  Glib::ustring::const_iterator it(text.begin());
  Glib::ustring::const_iterator end(text.end());

  int res = 0;

  while(it != end)
  {
    if (*it == c) ++res;
    ++it;
  }

  return res;
}

Glib::ustring
delete_trailing_separator(const Glib::ustring& path)
{
  Glib::ustring res(path);

  // Assure the path is in format "/home/name/some_dir".
  if ((! path.empty()) && (path[path.size() - 1] == '/'))
    res = path.substr(0, path.length() - 1);

  return res;
}

} // namespace String

namespace Time
{

/**
 * A convenience wrapper around strftime() and co.
 * Returns the time in 24-hour notation (%H:%M:%S).
 */
Glib::ustring
get_current_time()
{
  struct timeval tv;
  struct tm* ptm;
  char time_string[10];

  // Obtain the time of day, and convert it to a tm struct.
  gettimeofday(&tv, NULL);
  ptm = localtime(&tv.tv_sec);

  strftime(time_string, sizeof (time_string), "%T", ptm);
  Glib::ustring str(time_string);

  return str;
}

/**
 * Returns modification date in GTK+ style: Today/Yesterday/....
 */
Glib::ustring
get_modification_date_string(time_t timet)
{
  Glib::ustring date_str("Unknown");
  
  Glib::Date mtime;
  mtime.set_time(timet);
  
  time_t time_now;
  time_now = time(0);
  Glib::Date now;
  now.set_time(time_now);
	
  int days_diff = now.get_julian() - mtime.get_julian();
  Glib::ustring format;
  
  if (days_diff == 0)
  {
    date_str = "Today";
  }
  else if (days_diff == 1)
  {
    date_str = "Yesterday";
  }
  else
  {
    if (days_diff > 1 && days_diff < 7)
    {
      format = "%A"; // days from last week
    }
    else
    {
      format = "%x"; // any other date, use the locale-preferred representation
    }
    
    try
    {
      date_str = mtime.format_string(format);
    }
    catch (Glib::ConvertError ex)
    {
      g_warning("Cannot convert file modification time: %s",
		ex.what().c_str());
    }
  }
  
  if (date_str.empty())
  {
    date_str = "Unknown";
  }
  
  return date_str;
}

} // namespace Time

} // namespace Util

} // namespace USBSink
